/*

2sad 1.0
2sad 1.01 fix (only for DOS version)

by (c) Aley Keprt

*/


#ifdef WIN32
 #include "stdafx.h"
#endif

#ifdef DJGPP
 #include <dir.h>
 #include <dos.h>

 //this is 1.01 fix - remvoes wildcard expansion of DJGPP
 extern "C" char **__crt0_glob_function(char *_arg) {return 0;}
#endif


#include <stdio.h>
#include <stdlib.h>
//#include <io.h>
#include <string.h>

#include "2sad.h"
#include "zlib.h"


#define SAD_ID "Aley's disk backup"
#define SAD_ID_LENGTH (sizeof(SAD_ID)-1)


#define SKIP_SKIPPED "skipped"
#define SKIP_CANNOT_OPEN "cannot open!"
#define SKIP_CANNOT_READ "cannot read!"
#define SKIP_UNKNOWN_FORMAT "file format not recognized!"
#define SKIP_CANNOT_CREATE "cannot create!"
#define SKIP_WRITE_ERROR "write error!"


const char *errmsg[]={
 "",
 "Win32 system error",
 "Disk read error",
 "Out of memory",
};


const char *extension[]={".dsk",".sad",".sad"};


//----------------------------------------------------------------------------------


void xerr(TError err) {
 if(err!=ERR_None) printf("Error: %s\n",errmsg[err]);
 printf("\n");
 exit(err);
}


//----------------------------------------------------------------------------------


#define skipfile(text) if(1) {if(text[0]) printf("- " text "\n"); goto error;} else (void*)0

  
void convert_file(const char *fname,long fsize,TFormat of) {
//
//converts a single file
//input: file name,file size,requested file format
//
 FILE *f=0;
 gzFile gzf=0;
 char *data;
 
 char id[SAD_ID_LENGTH];

 struct {
  char sid,trk,sec,siz;
  long filesize;
 } form;


 for(;;) {

  //open file
  f=fopen(fname,"rb");
  if(!f) skipfile("");

  //read file
  if(!fread(id,SAD_ID_LENGTH,1,f)) skipfile("");
  if(!memcmp(id,SAD_ID,SAD_ID_LENGTH)) {

   //SAD
   if(of==FORM_SAD) skipfile("");

   if(!fread(&form,4,1,f)) skipfile("");
   form.filesize=form.sid*form.trk*form.sec*form.siz*64;
   printf("%s [SAD %iKB] ",fname,form.filesize/1024);
   fflush(stdout);

   data=new char[form.filesize];
   if(!data) {fclose(f);xerr(ERR_OutOfMemory);}

   if(!fread(data,form.filesize,1,f)) skipfile(SKIP_CANNOT_READ);
   fclose(f);f=0;

  } else if(id[0]==0x1f && id[1]==(char)0x8b) {
   
   //SAD.GZ
   if(of==FORM_SADGZ) skipfile("");
   fclose(f);f=0;

   //open gz
   gzFile gzf=gzopen(fname,"rb");
   if(!gzf) skipfile("");

   //check SAD id
   if(gzread(gzf,id,SAD_ID_LENGTH)!=SAD_ID_LENGTH) skipfile("");
   if(memcmp(id,SAD_ID,SAD_ID_LENGTH)) skipfile("");

   //read gz
   if(gzread(gzf,&form,4)!=4) skipfile("");
   form.filesize=form.sid*form.trk*form.sec*form.siz*64;
   printf("%s [SAD.GZ %iKB] ",fname,form.filesize/1024);
   fflush(stdout);

   data=new char[form.filesize];
   if(!data) {gzclose(gzf);xerr(ERR_OutOfMemory);}

   if(gzread(gzf,data,form.filesize)!=form.filesize) skipfile(SKIP_CANNOT_READ);
   gzclose(gzf);gzf=0;

  } else if(fsize==819200 || fsize==737280) {

   //DSK
   if(of==FORM_DSK) skipfile("");

   form.sid=2;
   form.trk=80;
   form.sec=(char)(fsize/81920);
   form.siz=512/64;
   form.filesize=fsize;
   printf("%s [DSK %iKB] ",fname,fsize/1024);
   fflush(stdout);

   char *dsk_data=new char[form.filesize];
   if(!dsk_data) {fclose(f);xerr(ERR_OutOfMemory);}

   rewind(f);
   if(!fread(dsk_data,form.filesize,1,f)) skipfile(SKIP_CANNOT_READ);
   fclose(f);f=0;

   //dsk --> sad
   data=new char[form.filesize];
   if(!data) xerr(ERR_OutOfMemory);

   int trksize=form.sec*512;
   char *dest=data;
   for(int side=0;side<2;side++) {
    for(int track=0;track<80;track++) {
     memcpy(dest,dsk_data+trksize*(2*track+side),trksize);
     dest+=trksize;
    }
   }

   free(dsk_data);

  } else {
   skipfile("");
  }

  //file is in the memory, we are going to save it

  //build new filename
  char name[260];
  strcpy(name,fname);
  int i=strlen(name);

  //find old extension
  for(;i>=0;i--) if(name[i]=='.') break;
  char *ext=name+i+1;

  //kill .gz
  if(!stricmp(ext,"gz")) {
   if(i>=0) name[i]=0;
   for(;i>=0;i--) if(name[i]=='.') break;
  }

  //kill old extension
  if(i>=0) name[i]=0;

  //append new extension
  strcat(name,extension[of]);

  printf("%s ",name);

  //save the file
  switch(of) {

  case FORM_DSK:
   f=fopen(name,"wb");
   if(!f) skipfile(SKIP_CANNOT_CREATE);

   {
   //sad --> dsk
   char *dsk_data=new char[form.filesize];
   if(!data) xerr(ERR_OutOfMemory);

   int trksize=form.sec*512;
   char *dest=dsk_data;
   for(int track=0;track<80;track++) {
    for(int side=0;side<2;side++) {
     memcpy(dest,data+trksize*(side*80+track),trksize);
     dest+=trksize;
    }
   }

   if(!fwrite(dsk_data,form.filesize,1,f)) skipfile(SKIP_WRITE_ERROR);
   fclose(f);f=0;

   free(dsk_data);
   }
   break;
   
  case FORM_SAD:
   f=fopen(name,"wb");
   if(!f) skipfile(SKIP_CANNOT_CREATE);

   if(fwrite(SAD_ID,1,SAD_ID_LENGTH,f)!=SAD_ID_LENGTH) skipfile(SKIP_WRITE_ERROR);
   if(fwrite(&form,1,4,f)!=4) skipfile(SKIP_WRITE_ERROR);

   if(fwrite(data,1,form.filesize,f)!=(size_t)form.filesize) skipfile(SKIP_WRITE_ERROR);
   fclose(f);f=0;

   break;

  case FORM_SADGZ:
   gzf=gzopen(name,"wb");
   if(!gzf) skipfile(SKIP_CANNOT_CREATE);
              
   if(gzwrite(gzf,SAD_ID,SAD_ID_LENGTH)!=SAD_ID_LENGTH) skipfile(SKIP_WRITE_ERROR);
   if(gzwrite(gzf,&form,4)!=4) skipfile(SKIP_WRITE_ERROR);

   if(gzwrite(gzf,data,form.filesize)!=form.filesize) skipfile(SKIP_WRITE_ERROR);
   gzclose(gzf);gzf=0;

   //show compression ratio
   f=fopen(name,"rb");
   if(f) {
    fseek(f,0,SEEK_END);
    printf("%i%%",100*ftell(f)/(form.filesize+22));
    fclose(f);f=0;
   }
   break;
  }

  free(data);
  printf("\n");
  return;
 }

 //error
error:
 if(f) {fclose(f);f=0;}
 if(gzf) {gzclose(gzf);gzf=0;}
}


void do_conversion(TFormat of,char *fname) {
//
//goes thru directory and processes any file matching the fname
//
#ifdef WIN32

 static struct _finddata_t fileinfo;

 long handle=_findfirst(fname,&fileinfo);
 if(handle==-1) {printf("Nothing to do!\n");return;}

 #define findnext() _findnext(handle,&fileinfo)

#else

 static struct ffblk fileinfo;

 if(findfirst(fname,&fileinfo,_A_ARCH|_A_RDONLY)) {
  printf("Nothing to do!\n");
  return;
 }

 #define findnext() findnext(&fileinfo)

 #define name ff_name
 #define size ff_fsize
 
#endif

 //save directory path
 char path[260];
 strcpy(path,fname);
 int i=strlen(path);

 for(;i>=0;i--) if(path[i]=='/' || path[i]=='\\' || path[i]==':') break;
 int pathlen=i+1;
 path[pathlen]=0;
 if(pathlen && path[pathlen-1]!='\\' && path[pathlen-1]!='/' && path[pathlen-1]!=':')
  path[pathlen++]='\\',path[pathlen]=0;

 //go thru files
 do {
#ifdef WIN32
  //skip directories
  if(fileinfo.attrib & _A_SUBDIR) continue;

#endif

  //paste name to the path
  strcpy(path+pathlen,fileinfo.name);

  //convert it
  convert_file(path,fileinfo.size,of);

 } while(!findnext());
}


//----------------------------------------------------------------------------------


void main(int argc,char *argv[]) {

 printf("2sad 1.0   (c) by Aley Keprt 1999\n"
        "=================================\n"
        "AleyKeprt@bigfoot.com, http://get.to/aley\n"
        "\n");


/*
     struct ffblk f;
     int done = findfirst("*.exe", &f, FA_ARCH|FA_RDONLY);
     while (!done)
     {
       printf("%10u %2u:%02u:%02u %2u/%02u/%4u %s\n",
         f.ff_fsize,
         (f.ff_ftime >> 11) & 0x1f,
         (f.ff_ftime >>  5) & 0x3f,
         (f.ff_ftime & 0x1f) * 2,
         (f.ff_fdate >>  5) & 0x0f,
         (f.ff_fdate & 0x1f),
         ((f.ff_fdate >> 9) & 0x7f) + 1980,
         f.ff_name);

       FILE *file=fopen(f.ff_name,"rb");
       getc(file);
       fclose(file);

       done = findnext(&f);
     }


 struct find_t f;

 if(!_dos_findfirst("*.*",_A_ARCH | _A_RDONLY, &f)) {
  do {
   printf("%-14s %10u %02u:%02u:%02u %02u/%02u/%04u\n",
                f.name,
                f.size,
                (f.wr_time >> 11) & 0x1f,
                (f.wr_time >>  5) & 0x3f,
                (f.wr_time & 0x1f) * 2,
                (f.wr_date >>  5) & 0x0f,
                (f.wr_date & 0x1f),
                ((f.wr_date >> 9) & 0x7f) + 1980);

   FILE *file=fopen(f.name,"rb");
   getc(file);
   fclose(file);

  } while(!_dos_findnext(&f));
 if(!_dos_findfirst("*.*",_A_ARCH | _A_RDONLY, &f)) {
 }
*/


 if(argc<=1) {
  printf("Portions Copyright (C) 1995-1998 Jean-loup Gailly and Mark Adler\n"
         "\n"
         "This utility converts floppy disk images between DSK and SAD formats.\n"
         "It offers full functionality of the former sad2dsk and dsk2sad.\n"
         "It is also able to convert between old and new SAD images (version 1 and 2).\n"
         "\n"
         "usage: 2sad [switch] <filename>\n"
         "\n"
         "switches: -d  DSK output (plain 720KB or 800KB files)\n"
         "          -s  plain SAD output (SAD version 1)\n"
         "          -g  GZip-packed SAD output (SAD version 2) [default]\n"
         "\n"
         "Filename can contain wildcards.\n"
         "\n"
         "examples: 2sad c:\\disks\\*.dsk   converts all DSK files in c:\\disks to SAD\n"
         "          2sad -d *   converts all files to DSK (when possible)\n"
         "\n");
  exit(0);
 }

 int namepos=1;

 //of=output format
 TFormat of=FORM_DEFAULT;

 if((argv[1][0]=='-' || argv[1][0]=='/') && argv[1][1] && !argv[1][2]) {
  switch(argv[1][1]|32) {

  case 'd':
   of=FORM_DSK;
   break;

  case 's':
   of=FORM_SAD;
   break;

  case 'g':
   of=FORM_SADGZ;
   break;

  default:
   namepos--;
  }
  namepos++;
 }

 if(argc<=namepos) printf("Nothing to do!\n\n");
 else do_conversion(of,argv[namepos]);
}

